from __future__ import annotations

from typing import TYPE_CHECKING, NamedTuple

from xsdata.formats.dataclass.filters import Filters
from xsdata.formats.dataclass.generator import DataclassGenerator

from ome_autogen import _util
from xsdata_pydantic_basemodel.generator import PydanticBaseFilters

if TYPE_CHECKING:
    from xsdata.codegen.models import Attr, Class
    from xsdata.codegen.resolver import DependenciesResolver
    from xsdata.models.config import GeneratorConfig


# from ome_types._mixins._base_type import AUTO_SEQUENCE
# avoiding import to avoid build-time dependency on the ome-types package
AUTO_SEQUENCE = "__auto_sequence__"


class Override(NamedTuple):
    element_name: str  # name of the attribute in the XSD
    class_name: str  # name of our override class
    module_name: str | None  # module where the override class is defined


CLASS_OVERRIDES = [
    Override("FillColor", "Color", "ome_types.model._color"),
    Override("StrokeColor", "Color", "ome_types.model._color"),
    Override("Color", "Color", "ome_types.model._color"),
    Override("Union", "ShapeUnion", "ome_types.model._shape_union"),
    Override(
        "StructuredAnnotations",
        "StructuredAnnotationList",
        "ome_types.model._structured_annotations",
    ),
]
# classes that should never be optional, but always have default_factories
NO_OPTIONAL = {"Union", "StructuredAnnotations"}

# if these names are found as default=..., turn them into default_factory=...
FACTORIZE = set([x.class_name for x in CLASS_OVERRIDES] + ["StructuredAnnotations"])

# prebuilt maps for usage in code below
OVERRIDE_ELEM_TO_CLASS = {o.element_name: o.class_name for o in CLASS_OVERRIDES}
IMPORT_PATTERNS = {
    o.module_name: {
        o.class_name: [f": {o.class_name} =", f": Optional[{o.class_name}] ="]
    }
    for o in CLASS_OVERRIDES
    if o.module_name
}


class OmeGenerator(DataclassGenerator):
    @classmethod
    def init_filters(cls, config: GeneratorConfig) -> Filters:
        return OmeFilters(config)

    def render_module(
        self, resolver: DependenciesResolver, classes: list[Class]
    ) -> str:
        mod = super().render_module(resolver, classes)

        # xsdata renames classes like "FillRule" (which appears as a SimpleType
        # inside of the Shape ComlexType) as "Shape_FillRule".
        # We want to make them available as "FillRule" in the corresponding
        # module, (i.e. the "Shape" module in this case).
        # That is, we want "Shape = Shape_FillRule" included in the module.
        # this is for backwards compatibility.
        aliases = []
        for c in classes:
            for i in resolver.imports:
                if f"{c.name}_" in i.qname:
                    # import_name is something like 'Shape_FillRule'
                    import_name = i.qname.rsplit("}", 1)[-1]
                    # desired alias is just 'FillRule'
                    alias = import_name.split(f"{c.name}_")[-1]
                    aliases.append(f"{alias} = {import_name}")

            # we also want inner (nested) classes to be available at the top level
            # e.g. map.Map.M -> map.M
            for inner in c.inner:
                aliases.append(f"{inner.name} = {c.name}.{inner.name}")

        if aliases:
            mod += "\n\n" + "\n".join(aliases) + "\n"

        return mod


class OmeFilters(PydanticBaseFilters):
    def __init__(self, config: GeneratorConfig):
        super().__init__(config)

        # TODO: it would be nice to know how to get the schema we're processing from
        # the config.  For now, we just assume it's the OME schema and that's the
        # hardcoded default in _util.get_appinfo
        self.appinfo = _util.get_appinfo()

    def class_bases(self, obj: Class, class_name: str) -> list[str]:
        # we don't need PydanticBaseFilters to add the Base class
        # because we add it in the config.extensions
        # This could go once PydanticBaseFilters is better about deduping
        return Filters.class_bases(self, obj, class_name)

    def _attr_is_optional(self, attr: Attr) -> bool:
        if attr.name in NO_OPTIONAL:
            return False
        return attr.is_nillable or (
            attr.default is None and (attr.is_optional or not self.format.kw_only)
        )

    def _format_type(self, attr: Attr, result: str) -> str:
        if self._attr_is_optional(attr):
            return f"None | {result}" if self.union_type else f"Optional[{result}]"
        return result

    def field_type(self, attr: Attr, parents: list[str]) -> str:
        if attr.is_list:
            # HACK
            # It would be nicer to put this in the self.field_name method...but that
            # method only receives the attr name, not the attr object, and so we
            # don't know at that point whether it belongs to a list or not.
            # This hack works only because this method is called BEFORE self.field_name
            # in the class.jinja2 template, so we directly modify the attr object here.
            attr.name = self.appinfo.plurals.get(attr.name, f"{attr.name}s")

        if attr.name in OVERRIDE_ELEM_TO_CLASS:
            return self._format_type(attr, OVERRIDE_ELEM_TO_CLASS[attr.name])

        type_name = super().field_type(attr, parents)
        # we want to use datetime.datetime instead of XmlDateTime
        return type_name.replace("XmlDateTime", "datetime")

    @classmethod
    def build_import_patterns(cls) -> dict[str, dict]:
        patterns = super().build_import_patterns()
        patterns.update(IMPORT_PATTERNS)
        patterns["ome_types._mixins._util"] = {"new_uuid": ["default_factory=new_uuid"]}
        patterns["datetime"] = {"datetime": ["datetime"]}
        return {key: patterns[key] for key in sorted(patterns)}

    def field_default_value(self, attr: Attr, ns_map: dict | None = None) -> str:
        if attr.tag == "Attribute" and attr.name == "ID":
            return repr(AUTO_SEQUENCE)
        for override in CLASS_OVERRIDES:
            if attr.name == override.element_name:
                if not self._attr_is_optional(attr):
                    return override.class_name
        return super().field_default_value(attr, ns_map)

    def format_arguments(self, kwargs: dict, indent: int = 0) -> str:
        # keep default_factory at the front
        if kwargs.get("default") in FACTORIZE:
            kwargs = {"default_factory": kwargs.pop("default"), **kwargs}

        # uncomment this to use new_uuid as the default_factory for all UUIDs
        # but then we have an equality checking problem in the tests
        # if kwargs.get("metadata", {}).get("pattern", "").startswith("(urn:uuid:"):
        #     kwargs.pop("default", None)
        #     kwargs = {"default_factory": "new_uuid", **kwargs}
        return super().format_arguments(kwargs, indent)

    def constant_name(self, name: str, class_name: str) -> str:
        if class_name in self.appinfo.enums:
            # use the enum names found in appinfo/xsdfu/enum
            return self.appinfo.enums[class_name][name].enum
        return super().constant_name(name, class_name)



from typing import Any, Dict

from pydantic import BaseModel


class KindMixin(BaseModel):
    def __init__(self, **data: Any) -> None:
        data.pop("kind", None)
        return super().__init__(**data)

    def dict(self, **kwargs: Any) -> Dict[str, Any]:
        d = super().dict(**kwargs)
        d["kind"] = self.__class__.__name__.lower()
        return d



from __future__ import annotations

import importlib
import operator
import os
import warnings
from dataclasses import is_dataclass
from pathlib import Path
from struct import Struct
from typing import TYPE_CHECKING, Any, cast

from pydantic import BaseModel

from ome_types.validation import validate_xml

try:
    from lxml import etree as ET
except ImportError:  # pragma: no cover
    from xml.etree import ElementTree as ET  # type: ignore[no-redef]

from xsdata.formats.dataclass.parsers.config import ParserConfig

from xsdata_pydantic_basemodel.bindings import (
    SerializerConfig,
    XmlParser,
    XmlSerializer,
)

if TYPE_CHECKING:
    import io
    from typing import TypedDict

    from xsdata.formats.dataclass.parsers.mixins import XmlHandler

    from ome_types._mixins._base_type import OMEType
    from ome_types.model import OME
    from xsdata_pydantic_basemodel.bindings import XmlContext

    class ParserKwargs(TypedDict, total=False):
        config: ParserConfig
        context: XmlContext
        handler: type[XmlHandler]


__all__ = ["from_xml", "to_xml", "to_dict", "from_tiff", "tiff2xml"]

OME_2016_06_URI = "http://www.openmicroscopy.org/Schemas/OME/2016-06"
MODULES = {
    OME_2016_06_URI: "ome_types._autogenerated.ome_2016_06",
}


def _get_ome_type(xml: str | bytes) -> type[OMEType]:
    """Resolve a python model class for the root element of an OME XML document."""
    if isinstance(xml, str) and not xml.startswith("<"):
        root = ET.parse(xml).getroot()  # noqa: S314
    else:
        if not isinstance(xml, bytes):
            xml = xml.encode("utf-8")
        root = ET.fromstring(xml)  # noqa: S314

    *_ns, localname = root.tag[1:].split("}", 1)
    ns = next(iter(_ns), None)

    if not ns or ns not in MODULES:
        raise ValueError(f"Unsupported OME schema tag {root.tag!r} in namespace {ns!r}")

    mod = importlib.import_module(MODULES[ns])
    try:
        return getattr(mod, localname)
    except AttributeError as e:  # pragma: no cover
        raise ValueError(
            f"Could not find a class for {localname!r} in {mod.__name__}"
        ) from e


def to_dict(source: OME | Path | str | bytes) -> dict[str, Any]:
    if is_dataclass(source):  # pragma: no cover
        raise NotImplementedError("dataclass -> dict is not supported yet")

    if isinstance(source, BaseModel):
        return source.dict(exclude_defaults=True)

    return from_xml(  # type: ignore[return-value]
        source,
        # the class_factory is what prevents class instantiation,
        # simply returning the params instead
        parser_kwargs={"config": ParserConfig(class_factory=lambda a, b: b)},
    )


def from_xml(
    xml: Path | str | bytes,
    *,
    validate: bool | None = None,
    parser: Any = None,
    parser_kwargs: ParserKwargs | None = None,
) -> OME:
    if parser is not None:  # pragma: no cover
        warnings.warn(
            "As of version 0.4.0, the parser argument is ignored. "
            "lxml will be used if available in the environment, but you can "
            "drop this keyword argument.",
            DeprecationWarning,
            stacklevel=2,
        )

    if validate:
        validate_xml(xml)

    if isinstance(xml, Path):
        xml = str(xml)

    # this cast is a lie... but it's by far the most common type that will
    # come out of this function, and will be more useful to most users.
    # For those who pass in an xml document that isn't just a root <OME> tag,
    # they can cast the result to the correct type themselves.
    OME_type = cast("type[OME]", _get_ome_type(xml))

    parser_ = XmlParser(**(parser_kwargs or {}))
    if isinstance(xml, bytes):
        return parser_.from_bytes(xml, OME_type)
    if os.path.isfile(xml):
        return parser_.parse(xml, OME_type)
    return parser_.from_string(xml, OME_type)


def to_xml(
    ome: OME,
    *,
    # exclude_defaults takes precendence over exclude_unset
    # if a value equals the default, it will be excluded
    exclude_defaults: bool = False,
    # exclude_unset will exclude any value that is not explicitly set
    # but will INCLUDE values that are set to their default
    exclude_unset: bool = True,
    indent: int = 2,
    include_namespace: bool | None = None,
    include_schema_location: bool = True,
    canonicalize: bool = False,
    validate: bool = False,
) -> str:
    config = SerializerConfig(
        pretty_print=(indent > 0) and not canonicalize,  # canonicalize does it for us
        pretty_print_indent=" " * indent,
        xml_declaration=False,
        ignore_default_attributes=exclude_defaults,
        ignore_unset_attributes=exclude_unset,
        attribute_sort_key=operator.attrgetter("name") if canonicalize else None,
    )
    if include_schema_location:
        config.schema_location = f"{OME_2016_06_URI} {OME_2016_06_URI}/ome.xsd"

    serializer = XmlSerializer(config=config)
    if include_namespace is None:
        include_namespace = canonicalize

    ns_map = {"ome" if include_namespace else None: OME_2016_06_URI}
    xml = serializer.render(ome, ns_map=ns_map)

    if canonicalize:
        xml = _canonicalize(xml, indent=" " * indent)
    if validate:
        validate_xml(xml)
    return xml


def _canonicalize(xml: str, indent: str) -> str:
    from xml.dom import minidom

    xml_out = ET.canonicalize(xml, strip_text=True)
    return minidom.parseString(xml_out).toprettyxml(indent=indent)  # noqa: S318


def from_tiff(
    path: Path | str,
    *,
    validate: bool | None = None,
    parser_kwargs: ParserKwargs | None = None,
) -> OME:
    xml = tiff2xml(path)
    return from_xml(xml, validate=validate, parser_kwargs=parser_kwargs)


TIFF_TYPES: dict[bytes, tuple[Struct, Struct, int, Struct]] = {
    b"II*\0": (Struct("<I"), Struct("<H"), 12, Struct("<H")),
    b"MM\0*": (Struct(">I"), Struct(">H"), 12, Struct(">H")),
    b"II+\0": (Struct("<Q"), Struct("<Q"), 20, Struct("<H")),
    b"MM\0+": (Struct(">Q"), Struct(">Q"), 20, Struct(">H")),
}


def _unpack(fh: io.BufferedReader, strct: Struct) -> int:
    return strct.unpack(fh.read(strct.size))[0]


def tiff2xml(path: Path | str) -> bytes:
    """Extract the OME-XML from a TIFF file."""
    with Path(path).open(mode="rb") as fh:
        head = fh.read(4)
        if head not in TIFF_TYPES:  # pragma: no cover
            raise ValueError(f"{path!r} does not have a recognized TIFF header")

        offset_fmt, tagno_fmt, tagsize, codeformat = TIFF_TYPES[head]
        offset_size = offset_fmt.size
        offset_size_4 = offset_size + 4

        if offset_size == 8:
            fh.seek(4, 1)
        fh.seek(_unpack(fh, offset_fmt))
        for _ in range(_unpack(fh, tagno_fmt)):
            tagstruct = fh.read(tagsize)
            if codeformat.unpack(tagstruct[:2])[0] == 270:
                size = offset_fmt.unpack(tagstruct[4:offset_size_4])[0]
                if size <= offset_size:
                    desc = tagstruct[offset_size_4 : offset_size_4 + size]
                    break
                fh.seek(offset_fmt.unpack(tagstruct[-offset_size:])[0])
                desc = fh.read(size)
                break
        else:  # pragma: no cover
            raise ValueError(f"No OME metadata found in file: {path}")
    if desc[-1] == 0:
        desc = desc[:-1]
    return desc



from __future__ import annotations

from typing import TYPE_CHECKING

from xsdata.codegen.analyzer import ClassAnalyzer
from xsdata.codegen.container import ClassContainer
from xsdata.codegen.handlers import RenameDuplicateAttributes
from xsdata.codegen.transformer import SchemaTransformer

if TYPE_CHECKING:
    from xsdata.codegen.models import Class


# we don't need RenameDuplicateAttributes because we inject
# proper enum names in our _generator.py
UNWANTED_HANDLERS = [(RenameDuplicateAttributes, None)]


class OMETransformer(SchemaTransformer):
    # overriding to remove the certain handlers
    def analyze_classes(self, classes: list[Class]) -> list[Class]:
        """Analyzer the given class list and simplify attributes and extensions."""
        # xsdata makes this particular class hard to extend/modify
        container = ClassContainer(config=self.config)

        for handlers in container.processors.values():
            for idx, h in enumerate(list(handlers)):
                for unwanted, wanted in UNWANTED_HANDLERS:
                    if isinstance(h, unwanted):
                        handlers.remove(h)
                        if wanted is not None:
                            handlers.insert(idx, wanted(container))

        container.extend(classes)

        return ClassAnalyzer.process(container)



from __future__ import annotations

import importlib.util
import sys
from importlib.abc import Loader, MetaPathFinder
from pathlib import Path
from typing import TYPE_CHECKING, Sequence

from ome_types._autogenerated.ome_2016_06 import *  # noqa

# these are here mostly to make mypy happy in pre-commit
# even when the model isn't built
from ome_types._autogenerated.ome_2016_06 import OME as OME
from ome_types._autogenerated.ome_2016_06 import Reference as Reference

if TYPE_CHECKING:
    from importlib.machinery import ModuleSpec
    from types import ModuleType

# ---------------------------------------------------------------------
# Below here is logic to allow importing from ome_types._autogenerated.ome_2016_06
# from ome_types.model.* (to preserve backwards compatibility)
# i.e. importing ome_types.model.map will import from
# ome_types._autogenerated.ome_2016_06.map ... and emit a warning


_OME_2016 = Path(__file__).parent.parent / "_autogenerated" / "ome_2016_06"


class OME2016Loader(Loader):
    def __init__(self, fullname: str) -> None:
        submodule = fullname.split(".", 2)[-1]
        file_2016 = (_OME_2016 / submodule.replace(".", "/")).with_suffix(".py")
        module_2016 = fullname.replace(".model.", "._autogenerated.ome_2016_06.", 1)
        if not file_2016.exists():  # pragma: no cover
            raise ImportError(
                f"Cannot find {submodule!r} in ome_types._autogenerated.ome_2016_06"
            )

        # warnings.warn(
        #     "Importing submodules from ome_types.model is deprecated. "
        #     "Please import types directly from ome_types.model instead.",
        #     stacklevel=2,
        # )
        self.fullname = fullname
        self.module_2016 = module_2016

    def create_module(self, spec: ModuleSpec) -> ModuleType | None:
        """Just return the 2016 version."""
        # this will already be in sys.modules because of the star import above
        return sys.modules[self.module_2016]

    def exec_module(self, module: ModuleType) -> None:
        """We never need to exec."""
        pass


# add a sys.meta_path hook to allow import of any modules in
# ome_types._autogenerated.ome_2016_06
class OMEMetaPathFinder(MetaPathFinder):
    def find_spec(
        self,
        fullname: str,
        path: Sequence[str] | None,
        target: ModuleType | None = None,
    ) -> ModuleSpec | None:
        """Return a module spec to redirect to ome_types._autogenerated.ome_2016_06."""
        if fullname.startswith("ome_types.model."):
            return importlib.util.spec_from_loader(fullname, OME2016Loader(fullname))
        return None


sys.meta_path.append(OMEMetaPathFinder())

from ome_types.model._converters import register_converters  # noqa

register_converters()
del register_converters



from __future__ import annotations

from typing import TYPE_CHECKING

from xsdata.formats.dataclass.filters import Filters
from xsdata.formats.dataclass.generator import DataclassGenerator
from xsdata.utils.collections import unique_sequence
from xsdata.utils.text import stop_words

if TYPE_CHECKING:
    from xsdata.codegen.models import Attr, Class
    from xsdata.models.config import GeneratorConfig, OutputFormat

stop_words.update(("schema", "validate"))


class PydanticBaseGenerator(DataclassGenerator):
    """Python pydantic dataclasses code generator."""

    @classmethod
    def init_filters(cls, config: GeneratorConfig) -> Filters:
        return PydanticBaseFilters(config)


class PydanticBaseFilters(Filters):
    @classmethod
    def build_import_patterns(cls) -> dict[str, dict]:
        patterns = Filters.build_import_patterns()
        patterns.update(
            {
                "pydantic": {
                    "Field": [" = Field("],
                    "BaseModel": ["BaseModel"],
                }
            }
        )
        return {key: patterns[key] for key in sorted(patterns)}

    @classmethod
    def build_class_annotation(cls, fmt: OutputFormat) -> str:
        # remove the @dataclass decorator
        return ""

    def field_definition(
        self, attr: Attr, ns_map: dict, parent_namespace: str | None, parents: list[str]
    ) -> str:
        defn = super().field_definition(attr, ns_map, parent_namespace, parents)
        return defn.replace("field(", "Field(")

    def format_arguments(self, kwargs: dict, indent: int = 0) -> str:
        # called by field_definition
        self.move_metadata_to_pydantic_field(kwargs)
        return super().format_arguments(kwargs, indent)

    def class_bases(self, obj: Class, class_name: str) -> list[str]:
        # add BaseModel to the class bases
        # FIXME ... need to dedupe superclasses
        bases = super().class_bases(obj, class_name)
        return unique_sequence([*bases, "BaseModel"])

    def move_metadata_to_pydantic_field(self, kwargs: dict, pop: bool = False) -> None:
        """Move metadata from the metadata dict to the pydantic Field kwargs."""
        # XXX: can we pop them?  or does xsdata need them in the metdata dict as well?
        if "metadata" not in kwargs:  # pragma: no cover
            return

        metadata: dict = kwargs["metadata"]
        getitem = metadata.pop if pop else metadata.get
        for from_, to_ in [
            ("min_inclusive", "ge"),
            ("min_exclusive", "gt"),
            ("max_inclusive", "le"),
            ("max_exclusive", "lt"),
            ("min_occurs", "min_items"),
            ("max_occurs", "max_items"),
            ("pattern", "regex"),
            ("min_length", "min_length"),
            ("max_length", "max_length"),
        ]:
            if from_ in metadata:
                kwargs[to_] = getitem(from_)



try:
    from importlib.metadata import PackageNotFoundError, version
except ImportError:
    from importlib_metadata import PackageNotFoundError, version  # type: ignore

try:
    __version__ = version("ome-types")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"

from ome_types import model
from ome_types._conversion import from_tiff, from_xml, to_dict, to_xml
from ome_types.model import OME
from ome_types.units import ureg
from ome_types.validation import validate_xml

__all__ = [
    "__version__",
    "from_tiff",
    "from_xml",
    "model",
    "OME",
    "to_dict",
    "to_xml",
    "ureg",
    "validate_xml",
]



from contextlib import suppress
from typing import Iterator, List

from pydantic import Field, ValidationError, validator

from ome_types._autogenerated.ome_2016_06.annotation import Annotation
from ome_types._autogenerated.ome_2016_06.boolean_annotation import BooleanAnnotation
from ome_types._autogenerated.ome_2016_06.comment_annotation import CommentAnnotation
from ome_types._autogenerated.ome_2016_06.double_annotation import DoubleAnnotation
from ome_types._autogenerated.ome_2016_06.file_annotation import FileAnnotation
from ome_types._autogenerated.ome_2016_06.list_annotation import ListAnnotation
from ome_types._autogenerated.ome_2016_06.long_annotation import LongAnnotation
from ome_types._autogenerated.ome_2016_06.map_annotation import MapAnnotation
from ome_types._autogenerated.ome_2016_06.tag_annotation import TagAnnotation
from ome_types._autogenerated.ome_2016_06.term_annotation import TermAnnotation
from ome_types._autogenerated.ome_2016_06.timestamp_annotation import (
    TimestampAnnotation,
)
from ome_types._autogenerated.ome_2016_06.xml_annotation import XMLAnnotation
from ome_types._mixins._base_type import OMEType
from ome_types.model._user_sequence import UserSequence

AnnotationTypes = (
    XMLAnnotation,
    FileAnnotation,
    ListAnnotation,
    LongAnnotation,
    DoubleAnnotation,
    CommentAnnotation,
    BooleanAnnotation,
    TimestampAnnotation,
    TagAnnotation,
    TermAnnotation,
    MapAnnotation,
)
_KINDS = {cls.__name__.lower(): cls for cls in AnnotationTypes}


class StructuredAnnotationList(OMEType, UserSequence[Annotation]):  # type: ignore[misc]
    # NOTE: in reality, this is List[StructuredAnnotationTypes]... but
    # for some reason that messes up xsdata data binding
    __root__: List[object] = Field(
        default_factory=list,
        metadata={
            "type": "Elements",
            "choices": tuple(
                {"name": cls.__name__, "type": cls} for cls in AnnotationTypes
            ),
        },
    )

    @validator("__root__", each_item=True)
    def _validate_root(cls, v: Annotation) -> Annotation:
        if isinstance(v, AnnotationTypes):
            return v
        if isinstance(v, dict):
            if "kind" in v:
                return _KINDS[v.pop("kind")](**v)
            for cls_ in AnnotationTypes:
                with suppress(ValidationError):
                    return cls_(**v)
        raise ValueError(f"Invalid Annotation: {v} of type {type(v)}")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__root__!r})"

    # overriding BaseModel.__iter__ to behave more like a real Sequence
    def __iter__(self) -> Iterator[Annotation]:  # type: ignore[override]
        yield from self.__root__  # type: ignore[misc]  # see NOTE above

    def __eq__(self, _value: object) -> bool:
        return _value == self.__root__



from __future__ import annotations

import io
import os
from contextlib import suppress
from functools import lru_cache
from pathlib import Path
from typing import IO, TYPE_CHECKING, Union

if TYPE_CHECKING:
    import xmlschema

    XMLSourceType = Union[str, bytes, Path, IO[str], IO[bytes]]

NS_OME = r"{http://www.openmicroscopy.org/Schemas/OME/2016-06}"
OME_2016_06_XSD = str(Path(__file__).parent / "ome-2016-06.xsd")


class ValidationError(ValueError):
    ...


def validate_xml(xml: XMLSourceType, schema: Path | str | None = None) -> None:
    """Validate XML against an XML Schema.

    By default, will validate against the OME 2016-06 schema.
    """
    with suppress(ImportError):
        return validate_xml_with_lxml(xml, schema)

    with suppress(ImportError):  # pragma: no cover
        return validate_xml_with_xmlschema(xml, schema)

    raise ImportError(  # pragma: no cover
        "Validation requires either `lxml` or `xmlschema`. "
        "Please pip install one of them."
    ) from None


def validate_xml_with_lxml(
    xml: XMLSourceType, schema: Path | str | None = None
) -> None:
    """Validate XML against an XML Schema using lxml."""
    from lxml import etree

    tree = etree.parse(schema or OME_2016_06_XSD)  # noqa: S320
    xmlschema = etree.XMLSchema(tree)

    if isinstance(xml, (str, bytes)) and not os.path.isfile(xml):
        xml = io.BytesIO(xml.encode("utf-8") if isinstance(xml, str) else xml)
        doc = etree.parse(xml)  # noqa: S320
    else:
        doc = etree.parse(xml)  # noqa: S320

    if not xmlschema.validate(doc):
        msg = f"Validation of {str(xml)[:20]!r} failed:"
        for error in xmlschema.error_log:
            msg += f"\n  - line {error.line}: {error.message}"
        raise ValidationError(msg)


def validate_xml_with_xmlschema(
    xml: XMLSourceType, schema: Path | str | None = None
) -> None:
    """Validate XML against an XML Schema using xmlschema."""
    from xmlschema.exceptions import XMLSchemaException

    xmlschema = _get_XMLSchema(schema or OME_2016_06_XSD)
    try:
        xmlschema.validate(xml)
    except XMLSchemaException as e:
        raise ValidationError(str(e)) from None


@lru_cache(maxsize=None)
def _get_XMLSchema(schema: Path | str) -> xmlschema.XMLSchema:
    import xmlschema

    xml_schema = xmlschema.XMLSchema(schema)
    # FIXME Hack to work around xmlschema poor support for keyrefs to
    # substitution groups
    ls_sgs = xml_schema.maps.substitution_groups[f"{NS_OME}LightSourceGroup"]
    ls_id_maps = xml_schema.maps.identities[f"{NS_OME}LightSourceIDKey"]
    ls_id_maps.elements = {e: None for e in ls_sgs}
    return xml_schema



from __future__ import annotations

import re
import warnings
from contextlib import suppress
from typing import TYPE_CHECKING, Any, cast

if TYPE_CHECKING:
    from pydantic import BaseModel
    from typing_extensions import Final

# Default value to support automatic numbering for id field values.
AUTO_SEQUENCE: Final = "__auto_sequence__"
# map of id_name -> max id value
ID_COUNTER: dict[str, int] = {}

# map of (id_name, id_value) -> converted id
# NOTE: this is cleared in OMEMixin.__init__, so that the set of converted IDs
# is unique to each OME instance
CONVERTED_IDS: dict[tuple[str, str], str] = {}


def _get_id_name_and_pattern(cls: type[BaseModel]) -> tuple[str, str]:
    # let this raise if it doesn't exist...
    # this should only be used on classes that have an id field
    id_field = cls.__fields__["id"]
    id_pattern = cast(str, id_field.field_info.regex)
    id_name = id_pattern.split(":")[-3]

    return id_name, id_pattern


def validate_id(cls: type[BaseModel], value: int | str) -> Any:
    """Pydantic validator for ID fields in OME models.

    This validator does the following:
    1. if it's valid string ID just use it, and updating the counter if necessary.
    2. if it's an invalid string id, try to extract the integer part from it, and use
       that to create a new ID, or use the next value in the sequence if not.
    2. if it's an integer, grab the appropriate ID name from the pattern and prepend it.
    3. if it's the special `AUTO_SEQUENCE` sentinel, use the next value in the sequence.

    COUNTERS stores the maximum previously-seen value on the class.
    """
    id_name, id_pattern = _get_id_name_and_pattern(cls)
    current_count = ID_COUNTER.setdefault(id_name, -1)

    if value == AUTO_SEQUENCE:
        # if it's the special sentinel, use the next value
        value = ID_COUNTER[id_name] + 1
    elif isinstance(value, str):
        if (id_name, value) in CONVERTED_IDS:
            # XXX: possible bug
            # if the same invalid value is used across multiple documents
            # we'll be replacing it with the same converted id here
            return CONVERTED_IDS[(id_name, value)]

        # if the value is a string, extract the number from it if possible
        value_id: str = value.rsplit(":", 1)[-1]

        # if the value matches the pattern, just return it
        # but update the counter if it's higher than the current value
        if re.match(id_pattern, value):
            with suppress(ValueError):
                # (not all IDs have integers after the colon)
                ID_COUNTER[id_name] = max(current_count, int(value_id))
            return value

        # if the value doesn't match the pattern, create a proper ID
        # (using the value_id as the integer part if possible)
        id_int = int(value_id) if value_id.isdecimal() else current_count + 1
        newname = validate_id(cls, id_int)
        # store the converted ID so we can use it elsewhere
        CONVERTED_IDS[(id_name, value)] = newname

        # warn the user
        msg = f"Casting invalid {id_name}ID {value!r} to {newname!r}"
        warnings.warn(msg, stacklevel=2)
        return newname
    elif not isinstance(value, int):
        raise ValueError(f"Invalid ID value: {value!r}, {type(value)}")

    # update the counter to be at least this value
    ID_COUNTER[id_name] = max(current_count, value)
    return f"{id_name}:{value}"



from __future__ import annotations

import warnings
import weakref
from typing import TYPE_CHECKING, Any, cast

from ome_types._mixins._base_type import OMEType
from ome_types._mixins._ids import CONVERTED_IDS

if TYPE_CHECKING:
    from pathlib import Path

    from ome_types._autogenerated.ome_2016_06 import OME, Reference


class OMEMixin:
    def __init__(self, **data: Any) -> None:
        # Clear the cache of converted IDs, so that they are unique to each OME instance
        CONVERTED_IDS.clear()
        super().__init__(**data)
        self._link_refs()

    def _link_refs(self) -> None:
        ids = collect_ids(self)
        for ref in collect_references(self):
            # all reference subclasses do actually have an 'id' field
            # but it's not declared in the base class
            if ref.id in ids:
                ref._ref = weakref.ref(ids[ref.id])
            else:
                warnings.warn(f"Reference to unknown ID: {ref.id}", stacklevel=2)

    def __setstate__(self, state: dict[str, Any]) -> None:
        """Support unpickle of our weakref references."""
        super().__setstate__(state)  # type: ignore
        self._link_refs()

    @classmethod
    def from_xml(cls, xml: Path | str) -> OME:
        from ome_types._conversion import from_xml

        return from_xml(xml)

    @classmethod
    def from_tiff(cls, path: Path | str) -> OME:
        from ome_types._conversion import from_tiff

        return from_tiff(path)

    def to_xml(self, **kwargs: Any) -> str:
        from ome_types._conversion import to_xml

        return to_xml(cast("OME", self), **kwargs)


def collect_ids(value: Any) -> dict[str, OMEType]:
    """Return a map of all model objects contained in value, keyed by id.

    Recursively walks all dataclass fields and iterates over lists. The base
    case is when value is neither a dataclass nor a list.
    """
    from ome_types.model import Reference

    ids: dict[str, OMEType] = {}
    if isinstance(value, list):
        for v in value:
            ids.update(collect_ids(v))
    elif isinstance(value, OMEType):
        for f in value.__fields__:
            if f == "id" and not isinstance(value, Reference):
                # We don't need to recurse on the id string, so just record it
                # and move on.
                ids[value.id] = value
            else:
                ids.update(collect_ids(getattr(value, f)))
    # Do nothing for uninteresting types.
    return ids


def collect_references(value: Any) -> list[Reference]:
    """Return a list of all References contained in value.

    Recursively walks all dataclass fields and iterates over lists. The base
    case is when value is either a Reference object, or an uninteresting type
    that we don't need to inspect further.

    """
    from ome_types.model import Reference

    references: list[Reference] = []
    if isinstance(value, Reference):
        references.append(value)
    elif isinstance(value, list):
        for v in value:
            references.extend(collect_references(v))
    elif isinstance(value, OMEType):
        for f in value.__fields__:
            references.extend(collect_references(getattr(value, f)))
    # Do nothing for uninteresting types
    return references



from pydantic import root_validator

from ome_types._mixins._base_type import OMEType


class PixelsMixin(OMEType):
    @root_validator(pre=True)
    def _validate_root(cls, values: dict) -> dict:
        if "metadata_only" in values:
            if isinstance(values["metadata_only"], bool):
                if not values["metadata_only"]:
                    values.pop("metadata_only")
                else:
                    # type ignore in case the autogeneration hasn't been built
                    from ome_types.model import MetadataOnly  # type: ignore

                    values["metadata_only"] = MetadataOnly()

        return values



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Callable, Generator, Iterator
from xml.etree.ElementTree import QName

from xsdata.formats.dataclass import context, parsers, serializers
from xsdata.formats.dataclass.serializers import config
from xsdata.formats.dataclass.serializers.mixins import XmlWriterEvent
from xsdata.models.enums import QNames
from xsdata.utils import collections, namespaces
from xsdata.utils.constants import EMPTY_MAP, return_input

if TYPE_CHECKING:
    from pydantic import BaseModel
    from xsdata.formats.dataclass.models.elements import XmlMeta


class XmlContext(context.XmlContext):
    """Pydantic BaseModel ready xml context instance."""

    def __init__(
        self,
        element_name_generator: Callable = return_input,
        attribute_name_generator: Callable = return_input,
    ):
        super().__init__(
            element_name_generator, attribute_name_generator, "pydantic-basemodel"
        )


class SerializerConfig(config.SerializerConfig):
    # here to add `ignore_unset_attributes` support to XmlSerializer
    __slots__ = (
        *config.SerializerConfig.__slots__,
        "ignore_unset_attributes",
        "attribute_sort_key",
    )

    def __init__(self, **kwargs: Any) -> None:
        self.ignore_unset_attributes = kwargs.pop("ignore_unset_attributes", False)
        self.attribute_sort_key = kwargs.pop("attribute_sort_key", None)
        super().__init__(**kwargs)


@dataclass
class XmlParser(parsers.XmlParser):
    context: XmlContext = field(default_factory=XmlContext)


@dataclass
class XmlSerializer(serializers.XmlSerializer):
    context: XmlContext = field(default_factory=XmlContext)

    # overriding so we can pass the args we want to next_attribute
    # and so that we can skip unset values
    def write_dataclass(
        self,
        obj: BaseModel,
        namespace: str | None = None,
        qname: str | None = None,
        nillable: bool = False,
        xsi_type: str | None = None,
    ) -> Generator:
        """
        Produce an events stream from a dataclass.

        Optionally override the qualified name and the xsi properties
        type and nil.
        """
        meta = self.context.build(
            obj.__class__, namespace, globalns=self.config.globalns
        )
        qname = qname or meta.qname
        nillable = nillable or meta.nillable
        namespace, tag = namespaces.split_qname(qname)

        yield XmlWriterEvent.START, qname

        # XXX: reason 1 for overriding.
        ignore_unset = getattr(self.config, "ignore_unset_attributes", False)
        for key, value in self.next_attribute(
            obj,
            meta,
            nillable,
            xsi_type,
            self.config.ignore_default_attributes,
            ignore_unset,
            getattr(self.config, "attribute_sort_key", None),
        ):
            yield XmlWriterEvent.ATTR, key, value

        for var, value in self.next_value(obj, meta):
            # XXX: reason 2 for overriding.
            if ignore_unset and var.name not in obj.__fields_set__:
                continue
            yield from self.write_value(value, var, namespace)

        yield XmlWriterEvent.END, qname

    # overriding so we can implement support for `ignore_unset_attributes`
    # and so that we can sort attributes as we want
    @classmethod
    def next_attribute(
        cls,
        obj: BaseModel,
        meta: XmlMeta,
        nillable: bool,
        xsi_type: str | None,
        ignore_optionals: bool,
        ignore_unset: bool = False,
        attribute_sort_key: Callable | None = None,
    ) -> Iterator[tuple[str, Any]]:
        """
        Return the attribute variables with their object values if set and not
        empty iterables.

        :param obj: Input object
        :param meta: Object metadata
        :param nillable: Is model nillable
        :param xsi_type: The true xsi:type of the object
        :param ignore_optionals: Skip optional attributes with default
            value
        :return:
        """

        set_fields = obj.__fields_set__ if ignore_unset else set()
        vars_ = meta.get_attribute_vars()
        if attribute_sort_key is not None:
            vars_ = sorted(meta.get_attribute_vars(), key=attribute_sort_key)

        # ^^^ new

        for var in vars_:
            if var.is_attribute:
                value = getattr(obj, var.name)
                if (
                    value is None
                    or (collections.is_array(value) and not value)
                    or (ignore_optionals and var.is_optional(value))
                    or (ignore_unset and var.name not in set_fields)  # new
                ):
                    continue

                yield var.qname, cls.encode(value, var)
            else:
                yield from getattr(obj, var.name, EMPTY_MAP).items()

        if xsi_type:
            yield QNames.XSI_TYPE, QName(xsi_type)

        if nillable:
            yield QNames.XSI_NIL, "true"


@dataclass
class JsonParser(parsers.JsonParser):
    context: XmlContext = field(default_factory=XmlContext)


@dataclass
class JsonSerializer(serializers.JsonSerializer):
    context: XmlContext = field(default_factory=XmlContext)


@dataclass
class UserXmlParser(parsers.UserXmlParser):
    context: XmlContext = field(default_factory=XmlContext)



import datetime
import warnings
from typing import Any

from xsdata.formats.converter import Converter, converter
from xsdata.models.datatype import XmlDateTime

from ome_types.model._color import Color


class DateTimeConverter(Converter):
    def serialize(self, value: datetime.datetime, **kwargs: Any) -> str:
        return str(XmlDateTime.from_datetime(value))

    def deserialize(self, value: Any, **kwargs: Any) -> datetime.datetime:
        xmldt = XmlDateTime.from_string(value)
        try:
            return xmldt.to_datetime()
        except ValueError as e:
            msg = f"Invalid datetime: {value!r} {e}."
            if xmldt.year <= 0:
                msg += "(BC dates are not supported)"
            warnings.warn(msg, stacklevel=2)
            return datetime.datetime(1, 1, 1)


class ColorConverter(Converter):
    def serialize(self, value: Color, **kwargs: Any) -> str:
        return str(value.as_int32())

    def deserialize(self, value: Any, **kwargs: Any) -> Color:
        return Color(value)


def register_converters() -> None:
    converter.register_converter(Color, ColorConverter())
    converter.register_converter(datetime.datetime, DateTimeConverter())



import warnings
from typing import Any, Dict

from pydantic import root_validator

from ome_types._mixins._base_type import OMEType


class BinDataMixin(OMEType):
    @root_validator(pre=True)
    def _v(cls, values: dict) -> Dict[str, Any]:
        # This catches the case of <BinData Length="0"/>, where the parser may have
        # omitted value from the dict, and sets value to b""
        # seems like it could be done in a default_factory, but that would
        # require more modification of xsdata I think
        if "value" not in values:
            if values.get("length") != 0:
                warnings.warn(
                    "BinData length is non-zero but value is missing", stacklevel=2
                )
            values["value"] = b""
        return values



from __future__ import annotations

from xsdata.codegen.writer import CodeWriter
from xsdata.models import config as cfg
from xsdata.utils import text

from ome_autogen._generator import OmeGenerator
from ome_autogen._util import camel_to_snake

KindedTypes = "(Shape|ManufacturerSpec|Annotation)"


MIXIN_MODULE = "ome_types._mixins"
MIXINS: list[tuple[str, str, bool]] = [
    (".*", f"{MIXIN_MODULE}._base_type.OMEType", False),  # base type on every class
    ("OME", f"{MIXIN_MODULE}._ome.OMEMixin", True),
    ("Instrument", f"{MIXIN_MODULE}._instrument.InstrumentMixin", False),
    ("Reference", f"{MIXIN_MODULE}._reference.ReferenceMixin", True),
    ("BinData", f"{MIXIN_MODULE}._bin_data.BinDataMixin", True),
    ("Pixels", f"{MIXIN_MODULE}._pixels.PixelsMixin", True),
    (KindedTypes, f"{MIXIN_MODULE}._kinded.KindMixin", True),
]

ALLOW_RESERVED_NAMES = {"type", "Type", "Union"}
OME_FORMAT = "OME"


def get_config(
    package: str, kw_only: bool = True, compound_fields: bool = False
) -> cfg.GeneratorConfig:
    # ALLOW "type" to be used as a field name
    text.stop_words.difference_update(ALLOW_RESERVED_NAMES)

    # use our own camel_to_snake
    # Our's interprets adjacent capital letters as two words
    # NameCase.SNAKE: 'PositionXUnit' -> 'position_xunit'
    # camel_to_snake: 'PositionXUnit' -> 'position_x_unit'
    cfg.__name_case_func__["snakeCase"] = camel_to_snake

    #  critical to be able to use the format="OME"
    CodeWriter.register_generator(OME_FORMAT, OmeGenerator)

    mixins = []
    for class_name, import_string, prepend in MIXINS:
        mixins.append(
            cfg.GeneratorExtension(
                type=cfg.ExtensionType.CLASS,
                class_name=class_name,
                import_string=import_string,
                prepend=prepend,
            )
        )

    keep_case = cfg.NameConvention(cfg.NameCase.ORIGINAL, "type")
    return cfg.GeneratorConfig(
        output=cfg.GeneratorOutput(
            package=package,
            # format.value lets us use our own generator
            # kw_only is important, it makes required fields actually be required
            format=cfg.OutputFormat(value=OME_FORMAT, kw_only=kw_only),
            structure_style=cfg.StructureStyle.CLUSTERS,
            docstring_style=cfg.DocstringStyle.NUMPY,
            compound_fields=cfg.CompoundFields(enabled=compound_fields),
        ),
        # Add our mixins
        extensions=cfg.GeneratorExtensions(mixins),
        # Don't convert things like XMLAnnotation to XmlAnnotation
        conventions=cfg.GeneratorConventions(class_name=keep_case),
    )



import warnings
from datetime import datetime
from enum import Enum
from textwrap import indent
from typing import TYPE_CHECKING, Any, ClassVar, Optional, Sequence, Set, Tuple, cast

from pydantic import BaseModel, validator

from ome_types._mixins._ids import validate_id
from ome_types.units import ureg

if TYPE_CHECKING:
    import pint


# Default value to support automatic numbering for id field values.
AUTO_SEQUENCE = "__auto_sequence__"


_UNIT_FIELD = "{}_unit"
_QUANTITY_FIELD = "{}_quantity"
DEPRECATED_NAMES = {
    "annotation_ref": "annotation_refs",
    "bin_data": "bin_data_blocks",
    "dataset_ref": "dataset_refs",
    "emission_filter_ref": "emission_filters",
    "excitation_filter_ref": "excitation_filters",
    "experimenter_ref": "experimenter_refs",
    "folder_ref": "folder_refs",
    "image_ref": "image_refs",
    "leader": "leaders",
    "light_source_settings": "light_source_settings_combinations",
    "m": "ms",
    "microbeam_manipulation_ref": "microbeam_manipulation_refs",
    "plate_ref": "plate_refs",
    "roi_ref": "roi_refs",
    "well_sample_ref": "well_sample_refs",
}


class OMEType(BaseModel):
    """The base class that all OME Types inherit from.

    This provides some global conveniences around auto-setting ids. (i.e., making them
    optional in the class constructor, but never ``None`` after initialization.).
    It provides a nice __repr__ that hides things that haven't been changed from
    defaults.  It adds ``*_quantity`` property for fields that have both a value and a
    unit, where ``*_quantity`` is a pint ``Quantity``.  It also provides pickling
    support.
    """

    # pydantic BaseModel configuration.
    # see: https://pydantic-docs.helpmanual.io/usage/model_config/
    class Config:
        arbitrary_types_allowed = False
        validate_assignment = True
        underscore_attrs_are_private = True
        use_enum_values = False
        validate_all = True

    # allow use with weakref
    __slots__: ClassVar[Set[str]] = {"__weakref__"}  # type: ignore

    _v = validator("id", pre=True, always=True, check_fields=False)(validate_id)

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        field_names = set(self.__fields__.keys())
        kwargs = set(data.keys())
        if kwargs - field_names:
            warnings.warn(
                f"Unrecognized fields for type {type(self)}: {kwargs - field_names}",
                stacklevel=2,
            )

    def __init_subclass__(cls) -> None:
        """Add `*_quantity` property for fields that have both a value and a unit.

        where `*_quantity` is a pint `Quantity`.
        """
        for field in cls.__fields__:
            if _UNIT_FIELD.format(field) in cls.__fields__:
                setattr(cls, _QUANTITY_FIELD.format(field), _quantity_property(field))

    def __repr_args__(self) -> Sequence[Tuple[Optional[str], Any]]:
        """Repr with only set values, and truncated sequences."""
        args = []
        for k, v in self._iter(exclude_defaults=True):
            if isinstance(v, Sequence) and not isinstance(v, str):
                # if this is a sequence with a long repr, just show the length
                # and type
                if len(repr(v).split(",")) > 5:
                    type_name = self.__fields__[k].type_.__name__
                    v = _RawRepr(f"[<{len(v)} {type_name}>]")
            elif isinstance(v, Enum):
                v = v.value
            elif isinstance(v, datetime):
                v = v.isoformat()
            args.append((k, v))
        return sorted(args, key=lambda f: f[0] not in ("name", "id"))

    def __repr__(self) -> str:
        lines = [f"{key}={val!r}," for key, val in self.__repr_args__()]
        if len(lines) == 1:
            body = lines[-1].rstrip(",")
        elif lines:
            body = "\n" + indent("\n".join(lines), "   ") + "\n"
        else:
            body = ""
        return f"{self.__class__.__qualname__}({body})"

    def __getattr__(self, key: str) -> Any:
        """Getattr that redirects deprecated names."""
        cls_name = self.__class__.__name__
        if key in DEPRECATED_NAMES and hasattr(self, DEPRECATED_NAMES[key]):
            new_key = DEPRECATED_NAMES[key]
            warnings.warn(
                f"Attribute '{cls_name}.{key}' is deprecated, use {new_key!r} instead",
                DeprecationWarning,
                stacklevel=2,
            )
            return getattr(self, new_key)
        raise AttributeError(f"{cls_name} object has no attribute {key!r}")


def _quantity_property(field_name: str) -> property:
    """Create property that returns a ``pint.Quantity`` combining value and unit."""

    def quantity(self: Any) -> Optional["pint.Quantity"]:
        value = getattr(self, field_name)
        if value is None:
            return None

        unit = cast("Enum", getattr(self, _UNIT_FIELD.format(field_name)))
        return ureg.Quantity(value, unit.value.replace(" ", "_"))

    return property(quantity)


class _RawRepr:
    """Helper class to allow repr to show raw values for fields that are sequences."""

    def __init__(self, raw: str) -> None:
        self.raw = raw

    def __repr__(self) -> str:
        return self.raw
